;;;	$Id: automata.lsp,v 1.18 2000/02/10 12:08:14 pawel Exp $	
;;; package AUTOMATA
;;; an implementation of the L* automata induction algorithm
;;; Copyleft 1998 Pawel Cichosz


(provide 'automata)
(require 'util)

(defpackage "AUTOMATA"
    (:documentation "Package providing an implementation of automata
induction algorithms, only L* so far.")
  (:use "COMMON-LISP" "UTIL")
  (:export "LEARN"))

(in-package "AUTOMATA")


(defvar *inalpha* '(0 1)
  "The default input alphabet.")

(defvar *outalpha* '(0 1)
  "The default output alphabet.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(declaim (inline learn))
(defun learn (&key (mq #'default-mq)
                   (eq #'default-eq)
                   (inalpha *inalpha*)
                   (outalpha *outalpha*))
  "Learn an automaton based on membership queries mq
and equivalence queries eq. Assume input alphabet inalpha and out alphabet
outalpha."
  (l*construct mq eq inalpha outalpha))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; The L* algorithm represents an automaton  indirectly by an L* table.
;;; The table consits of a set of row labels S and column labels E, as well
;;; as extended row labels S', all represented by lists.
;;; For each s in SuS' there is a table row, stored in a hash table,
;;; containing the automaton output when given string s.e in initial
;;; state for all e in E. We also redundantly store all answers to
;;; membership queries (automaton outputs) for all strings asked so far.
;;; This is to avoid re-asking the same queries.

(defstruct (l*table
             (:print-function l*table-print))
  "L* automaton table."
  (inalpha *inalpha* :type list)        ; input alphabet
  (outalpha *outalpha* :type list)      ; output alphabet
  (sset '(()) :type list)               ; row labels S
  (eset '(()) :type list)               ; column labels E
  (rows (make-hash-table :test #'equal) ; experiment results as rows
        :type hash-table)
  (entries (make-hash-table :test #'equal) ; and as single entries
           :type hash-table))


;;; The actual automaton is directly represented by its table, containg
;;; state transitions and outputs for all state-symbol pairs.

(defstruct (dfa
             (:print-function dfa-print))
  "Automaton table."
  (inalpha *inalpha* :type list)        ; input alphabet
  (outalpha *outalpha* :type list)      ; output alphabet
  (trans nil :type (vector (vector fixnum))) ; transition table
  (out nil :type (vector atom)))


;;; Some auxiliary stuff for examining and manipulating L* tables.


(declaim (inline l*row))
(defun l*row (lt s)
  "Get the row labeled by s from L* table lt."
  (gethash s (l*table-rows lt)))


(defsetf l*row l*row-set)

(declaim (inline l*row-set))
(defun l*row-set (lt s r)
  (setf (gethash s (l*table-rows lt)) r))


(declaim (inline l*row-equal))
(defun l*row-equal (lt r1 r2)
  "Check whether rows r1 and r2 are equal for the L* table lt. Return t if
they are or nil, e if they differ on experiment e."
  (let ((en (mismatch r1 r2 :from-end t)))
    (if en                              ; if r1 and r2 differ
        (values nil (nth (1- en) (l*table-eset lt))) ; return the experiment
        t)))                            ; no mistatch, rows equal


;;; L* tables are checked for consistency and closeness.


(defun l*consistent (lt)
  "Check whether the L* table lt is consistent. If yes, return t. Otherwise
return nil, s1, s2, a, e, where r(s1)=r(s2), but r(s1.a) and r(s1.a) differ
on e."
  (dolist (s1 (l*table-sset lt) t)      ; return t by default
    (let ((r1 (l*row lt s1)))           ; r(s1)
      (dolist (s2 (l*table-sset lt))
        (let ((r2 (l*row lt s2)))       ; r(s2)
          (when (l*row-equal lt r1 r2)  ; for any two equal rows
            (dolist (a (l*table-inalpha lt)) ; check all 1-symbol extensions
              (let ((r1_ (l*row lt (append s1 (list a)))) ; r(s1.a)
                    (r2_ (l*row lt (append s2 (list a))))) ; r(s2.a)
                (multiple-value-bind (q e)
                    (l*row-equal lt r1_ r2_) ; compare rows
                  (unless q             ; if not equal
                    (return-from l*consistent
                      (values nil s1 s2 a e))))))))))))
                                         

(defun l*closed (lt)
  "Check whether the L* table lt is closed. If yes, return t. Otherwise
return nil, s, a, where there is no s' in S such that r(s')=r(s.a)."
  (dolist (s (l*table-sset lt) t)       ; return t by default
    (dolist (a (l*table-inalpha lt))
      (let ((r (l*row lt (append s (list a))))) ; r(s.a)
        (when (notany #'(lambda (s)     ; when there is no equal row
                          (l*row-equal lt (l*row lt s) r))
                      (l*table-sset lt))
          (return-from l*closed         ; the table is not closed
            (values nil s a)))))))


;;; The L* table is extended by adding rows or columns and filling based
;;; on membership queries.


(defun l*extend-row (lt s mq)
  "Extend the L* table lt by adding row labeled by s, as well as
the corresponding extended rows. Fill the new entries by asking membership
queries mq. Return the modified table"
  (unless (l*row lt s)                  ; if no such row yet
    (l*fill-row lt s mq))               ; fill it
  (dolist (a (l*table-inalpha lt))
    (let ((s1 (append s (list a))))     ; for all possible extensions of s
      (unless (l*row lt s1)             ; if there is no corresponding row
        (l*fill-row lt s1 mq))))        ; fill it
  (pushnew s (l*table-sset lt) :test #'equal) ; add s to row labels
  lt)


(defun l*fill-row (lt s mq)
  "Fill the row labeled by s in the L* table lt, using membership query mq.
Return the modified table."
  (setf (l*row lt s)
        (mapcar #'(lambda (e)           ; get results for all experiments
                    (l*ask-membership lt (append s e) mq))
                (l*table-eset lt)))
  lt)


(defun l*extend-column (lt e mq)
  "Extend the L* table lt by adding column labeled by e. Fill the new entries
by asking membership queries mq. Return the modified table."
  (push e (l*table-eset lt))
  (maphash #'(lambda (s r)              ; new experiment result
               (push (l*ask-membership lt (append s e) mq) r)
               (setf (l*row lt s) r))   ; this is OK
           (l*table-rows lt))
  lt)


;;; Table and consistent L* tables are translated into automata.


(defun l*->dfa (lt)
  "Construct an automaton based on the L* table lt."
  (let ((states (make-hash-table :test #'equal)) ; map rows to states
        (i 1))                          ; state number
    (setf (gethash (l*row lt '()) states) 0) ; initial state
    (dolist (s (l*table-sset lt))
      (let ((r (l*row lt s)))           ; r(s)
        (unless (gethash r states)      ; using rows as keys
          (setf (gethash r states) i)   ; (new) state number for row r
          (incf i))))                   ; next number
    (make-dfa :inalpha (l*table-inalpha lt)
              :outalpha (l*table-outalpha lt)
              :trans                    ; fill the transition table
              (let ((trans (make-array (hash-table-count states))))
                (dolist (s (l*table-sset lt) trans) ; too many iterations???
                  (let* ((r (l*row lt s))
                         (i (gethash r states)))
                    (setf (svref trans i)
                          (map 'vector  ; transition vector for state #i
                               #'(lambda (a)
                                   (gethash (l*row lt (append s (list a)))
                                            states))
                               (l*table-inalpha lt))))))
              :out (let ((out (make-array (hash-table-count states))))
                     (maphash #'(lambda (r i)
                                  (setf (svref out i) ; output is result for
                                        (first (last r)))) ; null experiment
                              states)
                     out))))


;;; The L* algorithm consists in finding a consistent and closed table using
;;; membership queries. This is the single L* iteration. Several such
;;; iterations may be necessary until the equivalence query on
;;; the auotmaton corresponding to the closed and consistent table receives
;;; the positive answer.


(defun l*iteration (lt mq)
  "Make a single L* iteration using the initial L* table lt and membership
queries mq. Return the final, closed and consistent table."
  (do* ((con-list (multiple-value-list (l*consistent lt))
                  (multiple-value-list (l*consistent lt)))
        (clo-list (multiple-value-list (l*closed lt))
                  (multiple-value-list (l*closed lt)))
        (is-cons (first con-list)       ; is lt consistent
                 (first con-list))
        (is-clo (first clo-list)        ; is lt closed
                (first clo-list)))
       ((and is-cons is-clo) lt)        ; terminate when consistent and closed
    (cond ((not is-cons)                ; if lt is not consistent
           (let ((s1 (second con-list))
                 (s2 (third con-list))
                 (a (fourth con-list))
                 (e (fifth con-list)))
             (l*extend-column lt (cons a e) mq)))
          ((not is-clo)                 ; if lt is not closed
           (let ((s (second clo-list))
                 (a (third clo-list)))
             (l*extend-row lt (append s (list a)) mq))))))


(defun l*construct (mq eq inalpha outalpha)
  "Construct a dfa based on membership queries mq and equivalence
queries eq, using input alphabet inalpha and output alphabet outalpha."
  (do* ((lt (l*extend-row (make-l*table :inalpha inalpha
                                        :outalpha outalpha) ; new L* table
                          '() mq)       ; and its initial row
            (l*iteration lt mq))
        (auto nil                       ; the automaton corresponding to lt
              (l*->dfa lt))
        (eq-list nil                    ; equivalence query results
                 (multiple-value-list (l*ask-equivalence auto eq)))
        (is-eq nil                      ; is auto equivalent to the target
               (first eq-list)))
       (is-eq auto)                     ; terminate when equivalent
    (let ((cntex (second eq-list)))     ; the counterexample
      (dolist (s (prefixes cntex))      ; for all prefixes
        (l*extend-row lt s mq)))))


;;; The L* algorithm may ask membership and equivalence queries.


(defun l*ask-membership (lt s mq)
  "Ask the membership query mq for the L* table lt and string s or restore
the previous answer, it it has been asked already."
  (or (gethash s (l*table-entries lt))
      (setf (gethash s (l*table-entries lt)) (funcall mq s))))


(defun l*ask-equivalence (auto eq)
  "Ask the equivalence query eq for automaton auto."
  (funcall eq auto))


;;; The default membership and equivalence queries ask the user and
;;; read its response.


(declaim (inline default-mq))
(defun default-mq (s)
  "Ask a membership query for string a. The user is expected to type
a valid output symbol."
  (format t "~&Membership query for ~A~&" s)
  (read))


(declaim (inline default-eq))
(defun default-eq (auto)
  "Ask an equivalence query for automaton auto. The user is expected to type
a list containg a true false indicator and, when false, a couterexample."
  (format t "~&Equivalence query for ~A~&" auto)
  (values-list (read)))


;;; The printing functions for L* tables and automatons.


(defun l*table-print (lt stream depth)
  "Print the L* table lt to stream in a possibly human readable form."
  (declare (ignore depth))              ; not using this argument
  (format stream "~&A=~A" (l*table-inalpha lt))
  (format stream "~&Y=~A" (l*table-outalpha lt))
  (format stream "~&S=~A" (l*table-sset lt))
  (format stream "~&E=~A" (l*table-eset lt))
  (format stream "~&Regular rows:")
  (dolist (s (l*table-sset lt))
    (format stream "~&r(~A) = ~A" s (l*row lt s)))
  (format stream "~&Extended rows:")
  (maphash #'(lambda (s r)
               (unless (member s (l*table-sset lt) :test #'equal)
                 (format stream "~&r(~A) = ~A" s r)))
           (l*table-rows lt)))


(defun dfa-print (auto stream depth)
  "Print the automaton auto to stream in a possibly human readable form."
  (declare (ignore depth))              ; not using this argument
  (format stream "~&A=~A" (dfa-inalpha auto))
  (format stream "~&Y=~A" (dfa-outalpha auto))
  (dotimes (i (length (dfa-trans auto)))
    (format stream "~&~A: ~A | ~A"
            i (svref (dfa-trans auto) i) (svref (dfa-out auto) i))))
