;;;	$Id: bayes.lsp,v 1.35 2000/02/10 12:06:58 pawel Exp $
;;; package BAYES
;;; a simple Common Lisp implementation of naive Bayes classifier
;;; Copyleft 1998 Pawel Cichosz

(provide 'bayes)
(require 'concept)
(require 'util)

(defpackage "BAYES"
    (:documentation "Package providing an implementation of naive Bayes
classifier.")
  (:use "COMMON-LISP" "CONCEPT" "UTIL")
  (:export "*N-VIRTUAL*"
           "LEARN"
           "CLASSIFY"
           "HYPOTHESIS-SIZE"))

(in-package "BAYES")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defparameter *epsilon* 1/10000
  "The value assumed for close to zero probabilities. Not actually used.")


(declaim (inline learn)
         (ftype (function (list &key) probestimate) learn))
(defun learn (instances &key)
  "Learn from instances assuming the domain described by
attributes and categories."
  (probestimate-compute instances))


(declaim (inline classify)
         (ftype (function (probestimate cons &key) atom) classify))
(defun classify (probabilities i &key)
  "Classify instance i using probabilities."
  (category-at (argmax (bayes-probabilities probabilities i))))


(declaim (inline hypothesis-size))
(defun hypothesis-size (probabilities &key)
  "Return the number of category probabilities as the first value
and the number of attribute-value probabilities per category as
the second value."
  (values (length (probestimate-cestim probabilities))
          (aestim-size (probestimate-aestim probabilities))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Hypotheses of naive Bayes classifier are represented by
;;; probability estimates.


(defstruct (probestimate
             (:print-function probestimate-print))
  (cestim nil :type (vector (real 0 1))) ; category estimates
  (aestim nil :type (array (vector (real 0 1)) (* *)))) ; attribute estimates


(declaim (inline probestimate-compute)
         (ftype (function (list) probestimate) probestimate-compute))
(defun probestimate-compute (instances)
  "Compute probability estimates for instances. Produces a pair the first
element of which is a vector of category probability estimates and
the second element of which is a 2-dimensional array of vectors
of attribute probability estimates."
  (make-probestimate :cestim (category-estimates instances)
                     :aestim (attribute-estimates instances)))


;;; Category probability estimates are kept in an array that stores
;;; the estimated probability of each category. For attributes, there
;;; is a 2-dimensional array indexed first by category number and then
;;; by attribute number, that stores vectors of probabilities for
;;; successive values of the corresponding attribute given the
;;; corresponding category.
;;; Incidentally, the below two functions are somewhat inefficient
;;; (they iterate over categories and attributes rather than over
;;; instances, they duplicate some computation), but this is in order
;;; to formulate them as mapping over the result produced by
;;; category-distribution and value-distribution-c.


(declaim (ftype (function (list) (vector (real 0 1))) category-estimates))
(defun category-estimates (instances)
  "Compute the 1-dimensional array of category probability estimates."
  (let ((ni (length instances)))
    (map 'vector                        ; for all categories
	 #'(lambda (cn)                 ; given count estimate probability
	     (m-estimate cn ni (categories-length)))
	 (category-distribution instances))))


(declaim (ftype (function (list) (array (vector (real 0 1)) (* *)))
                attribute-estimates))
(defun attribute-estimates (instances)
  "Compute the 2-dimensional array of vectors of conditional attribute
probability estimates given categories."
  (let ((p (make-array `(,(categories-length) ; array of probability
                         ,(attributes-length)) ; vectors
		       :element-type '(vector (real 0 1))))) ; to be returned
    (do-categories-n (c cn p)                ; for all categories
      (let ((nic (instances-count-c instances c))) ; # instances in category c
        (do-attributes-n (a an)              ; for all attributes
          (setf (aref p cn an)          ; for category #cn and attribute #an
                (map 'vector            ; generate probabilities vector
                     #'(lambda (vn)     ; for all attribute values
                         (m-estimate vn nic (attribute-values-length a)))
                     (value-distribution-c instances a c))))))))


(defun aestim-size (ae)
  (declare (ignore ae))                 ; argument not needed
  (let ((as 0))
    (do-attributes (a as)
      (incf as (attribute-values-length a)))))


;;; Probabilities are estimated based on instance distribution
;;; using m-estimate.


(declaim (inline m-estimate)
         (ftype (function (fixnum fixnum &optional fixnum (real 0 1))
                          (real 0 1))
                m-estimate))
(defun m-estimate (n1 n &optional
                      (m 1)             ; number of virtual instances
                      (p (if (> m 0) (/ m) 0))) ; prior probability
  "Estimate the probability of an event observed n1 times out of n total
events in the training set, assuming its prior probability p."
  (if (or (> m 0) (> n 0))              ; if at least one instance
      (/ (+ n1 (* m p)) (+ n m))        ; compute probability
      *epsilon*))                       ; otherwise use default


;;; The real Bayesian stuff comes here. Actually, a more reasonable
;;; implementation would sum logarithms of probabilities rather than
;;; multiply probabilities, but direct manipulating probabilities has
;;; been preferred for didactic reasons.


(declaim (ftype (function (probestimate cons) sequence) bayes-probabilities))
(defun bayes-probabilities (probabilities i)
  "Get the sequence with Bayes estimates of category probabilities for
instance i."
  (let ((cp (probestimate-cestim probabilities)) ; category probabilities
	(ap (probestimate-aestim probabilities)) ; attribute probabilities
	(cn 0))                         ; current category number
    (map 'vector
         #'(lambda (c)
             (declare (ignore c))       ; we don't care for the argument
             (prog1
                 (* (svref cp cn)
                    (attributes-probability ap i cn))
               (incf cn)))              ; next category number
         (categories-sequence))))


(declaim (ftype (function ((array (vector (real 0 1)) (* *)) cons fixnum)
                          (real 0 1)) attributes-probability))
(defun attributes-probability (ap i cn)
  "Get the probability of attribute values of instance i assuming
category number cn, given conditional attribute probability estimates ap."
  (let ((an 0))                         ; current attribute number
    (reduce #'*                         ; multiply all
	    (instance-values i)         ; (hopefully) independent attribute
	    :key #'(lambda (v)          ; probabilities
                     (prog1
                         (svref (aref ap cn an)
                               (attribute-value-position v (attribute-at an)))
		       (incf an))))))   ; next attribute number


;;; Printing.


(defun probestimate-print (probabilities stream depth)
  "Print the probability estimates probabilities."
  (declare (ignore depth))              ; not using this argument
  (format stream "~&Category probability estimates:~%")
  (do-categories-n (c cn)
    (format stream
            "~&Pr(~A)=~A"
            c
            (svref (probestimate-cestim probabilities) cn)))
  (format stream "~&Attribute conditional probability estimates:~%")
  (do-categories-n (c cn)                  ; for all categories
    (do-attributes-n (a an)              ; for all attributes
      (do-attribute-values-n (v vn a)  ; for all attribute values
        (format stream
                "~&Pr(~A=~A|~A)=~A"
                (attribute-name a) v c
                (svref (aref (probestimate-aestim probabilities) cn an)
                      vn))))))
