;;;	$Id: cluster.lsp,v 1.10 2000/02/10 12:03:56 pawel Exp $	
;;; package CLUSTER
;;; a simplified implementation of a COBWEB-like clustering algorithm
;;; Copyleft 1998 Pawel Cichosz

(provide 'cluster)
(require 'concept)

(defpackage "CLUSTER"
    (:documentation "Package providing an implementation of the COBWEB
clustering algorithm.")
  (:use "COMMON-LISP" "CONCEPT" "UTIL")
  (:export "LEARN"
           "CLASSIFY"
           "HYPOTHESIS-SIZE"))

(in-package "CLUSTER")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(declaim (inline learn)
         (ftype (function (list &key) clustree) learn))
(defun learn (instances &key)
  "Learn a clustering for the given set of instances."
  (clustree-build instances))


;;; Classification relies in finding a 1-level host for the given instance
;;; and returning its majority label, if any.

(declaim (inline classify)
         (ftype (function (clustree cons &key) atom) classify))
(defun classify (ctree i &key)
  "Classify instance i to the majority category of its best host among
the direct descendants of clustering tree ctree."
  (let ((hosts (clustree-hosts ctree i)))
    (category-at (argmax (clustree-cdistr (first (first hosts)))))))


(defun hypothesis-size (ctree &key)
  "Return the total number of nodes of clustering tree ctree as the first
value and the number of 1-level nodes as the second value."
  (cond ((null ctree)
         (values 0 0))
        (t (values (1+ (reduce #'+ (mapcar #'hypothesis-size
                                           (clustree-subtrees ctree))))
                   (length (clustree-subtrees ctree))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Clustering is represented by a clustering tree. Each node contains
;;; attribute value distribution and category label distribution for
;;; instances hosted by this node, if using for supervised learning.
;;; The category distribution is only used for for instance classification
;;; based on majority category and for printing clustering trees.


(defstruct (clustree
             (:print-function clustree-print))
  "Clustering tree structure."
  (card 0 :type fixnum)                 ; number of hosted instances
  (adistr nil :type (vector (vector fixnum))) ; attribute value distribution
  (cdistr nil :type (vector fixnum))    ; category distribution
  (subtrees nil :type list))


(declaim (ftype (function (list) clustree) clustree-build))
(defun clustree-build (instances)
  "Build clustering tree for the given set of instances."
  (let ((ctree nil))
    (dolist (i instances ctree)
      (setq ctree (clustree-add ctree i)))))


;;; Here comes the top-level code for the basic COBWEB algorithm.


(declaim (ftype (function (clustree cons) clustree) clustree-add))
(defun clustree-add (ctree i)
  "Add instance i to clustering tree ctree using the basic COBWEB algorithm."
  (cond ((null ctree)                   ; if adding the first instance
         (clustree-make-single i))      ; make a leaf
        ((= (clustree-card ctree) 1)    ; if this is a leaf
         (clustree-adjoin ctree (clustree-make-single i)))
        (t                              ; otherwise
         (clustree-add1 ctree i)        ; add to this very node
         ;; compute utilities of possible operators
         (let* ((hosts (clustree-hosts ctree i))
                (best (first hosts))    ; best host
                (best2 (second hosts))  ; second best host
                (usingle (clustree-try-single ctree i)) ; single leaf utility
                (ubest (rest best)) ; best host utility
                (umerge (clustree-try-merge ctree
                                            (first best)
                                            (first best2) i)) ; merge utility
                (usplit (clustree-try-split ctree
                                            (first best) i)) ; split utility
                (umax (funcall #'max usingle ubest umerge usplit)))
           ;; and choose the best operator to actually perform
           (cond ((= umax usingle)
                  (clustree-single ctree i))
                 ((= umax ubest)
                  (clustree-add (first best) i)) ; direct recursion
                 ((= umax umerge)
                  (clustree-merge ctree (first best) (first best2) i))
                 ((= umax usplit)
                  (clustree-split ctree (first best) i))))
         ctree)))


;;; And now some more low-level details.


;;; Shallow (non-recursive) add/remove.

(declaim (ftype (function (clustree cons) clustree) clustree-add1))
(defun clustree-add1 (ctree i)
  "Add instance i to the root of clustering tree ctree, without descending
to subtrees. Return the modified tree."
  (incf (clustree-card ctree))          ; one more instance
  (let ((an 0))                         ; current attribute number
    (map nil                            ; increase appropriate value counts
         #'(lambda (v a)                ; for value v of attribute a
             (incf (svref (svref (clustree-adistr ctree) an)
                         (attribute-value-position v a)))
             (incf an))                 ; next attribute number
         (instance-values i)
         (attributes-sequence)))
  (incf (svref (clustree-cdistr ctree)   ; increase category count
              (category-position (instance-category i))))
  ctree)


(declaim (ftype (function (clustree cons) clustree) clustree-remove1))
(defun clustree-remove1 (ctree i)
  "Remove instance i from the root of clustering tree ctree, without
descending to subtrees. Return the modified tree."
  (decf (clustree-card ctree))          ; one more instance
  (let ((an 0))                         ; current attribute number
    (map nil                            ; decrease appropriate value counts
         #'(lambda (v a)                ; for value v of attribute a
             (decf (svref (svref (clustree-adistr ctree) an)
                         (attribute-value-position v a)))
             (incf an))                 ; next attribute number
         (instance-values i)
         (attributes-sequence)))
  (decf (svref (clustree-cdistr ctree)   ; decrease category count
              (category-position (instance-category i))))
  ctree)


;;; Functions for trying and actually performing operators. One source
;;; of inefficiency is that the actually selected operator performs
;;; again the computation perfomed earlier when it was tried.


(declaim (ftype (function (clustree cons) real) clustree-try-single))
(defun clustree-try-single (ctree i)
  "Temporarily try to place instance i in a leaf descendant of ctree
and return the utility of the resulting clustering."
  (push (clustree-make-single i) (clustree-subtrees ctree)) ; add leaf
  (prog1
      (clustree-utility ctree)          ; compute utility
    (setf (clustree-subtrees ctree)     ; remove leaf
          (rest (clustree-subtrees ctree)))))


(declaim (ftype (function (clustree cons) clustree) clustree-single))
(defun clustree-single (ctree i)
  "Place instance i in a leaf descendant of clustering tree ctree
and return of the modified clustering tree."
  (push (clustree-make-single i) (clustree-subtrees ctree)) ; add leaf
  ctree)


(declaim (ftype (function (clustree cons) real) clustree-try-add))
(defun clustree-try-add (ctree h i)
  "Temporarily try to place instance i in descendant h of ctree
and return the utility of the resulting clustering."
  (clustree-add1 h i)                   ; add i to node h
  (prog1
      (clustree-utility ctree)          ; compute utility
    (clustree-remove1 h i)))            ; and remove i from h


(declaim (ftype (function (clustree clustree clustree cons) real)
                clustree-try-merge))
(defun clustree-try-merge (ctree h1 h2 i)
  "Temporarily try to merge h1 and h2 the descendants of ctree, place
instance i in the merged node, and return the utility of the resulting
clustering."
  (let ((old-subtrees (copy-list (clustree-subtrees ctree))))
    (setf (clustree-subtrees ctree)
          (delete h1 (clustree-subtrees ctree)))
    (setf (clustree-subtrees ctree)
          (delete h2 (clustree-subtrees ctree)))
    (push (clustree-add1 (clustree-make-joint h1 h2) i)
          (clustree-subtrees ctree))
    (prog1
        (clustree-utility ctree)
      (setf (clustree-subtrees ctree) old-subtrees))))


(declaim (ftype (function (clustree clustree clustree cons) clustree)
                clustree-merge))
(defun clustree-merge (ctree h1 h2 i)
  "Merge descendants of clustering tree ctree, h1 and h2, place instance i
in the merged node, and return the modified clustering tree."
  (setf (clustree-subtrees ctree)
        (delete h1 (clustree-subtrees ctree)))
  (setf (clustree-subtrees ctree)
        (delete h2 (clustree-subtrees ctree)))
  (push (clustree-add (clustree-make-joint h1 h2) i) ; indirect recursion
        (clustree-subtrees ctree))
  ctree)


(declaim (ftype (function (clustree clustree cons) real) clustree-try-split))
(defun clustree-try-split (ctree h i)
  "Temporarily try to split h the descendant of ctree, place instance i
in the best host among its descendants, and return the resulting clustering
utility."
  (let ((old-subtrees (copy-list (clustree-subtrees ctree))))
    (setf (clustree-subtrees ctree)
          (delete h (clustree-subtrees ctree))) ; delete the node to split
    (dolist (s (clustree-subtrees h))
      (push s (clustree-subtrees ctree))) ; add its subtrees
    (prog1                              ; not sure if this is really COBWEB
        (let ((umax (loop for s in (clustree-subtrees h)
                          maximize (clustree-try-add ctree s i))))
          (or umax most-negative-single-float)) ; check for null umax
      (setf (clustree-subtrees ctree) old-subtrees))))


(declaim (ftype (function (clustree clustree cons) clustree) clustree-split))
(defun clustree-split (ctree h i)
  "Split h the descendant of clustering tree ctree, place instance i
in the best host among its descendants, and return the modified clustering
tree."
  (setf (clustree-subtrees ctree)
        (delete h (clustree-subtrees ctree))) ; delete the node to split
  (dolist (s (clustree-subtrees h))
    (push s (clustree-subtrees ctree))) ; add its subtrees
  (clustree-add (clustree-remove1 ctree i) i)) ; indirect recursion


(declaim (inline clustree-make-single)
         (ftype (function (cons) clustree) clustree-make-single))
(defun clustree-make-single (i)
  "Make a leaf clustering tree for instance i."
  (make-clustree :card 1                ; make a new leaf
                 :adistr (adistr-single i)
                 :cdistr (cdistr-single i)
                 :subtrees nil))


(declaim (inline clustree-make-joint)
         (ftype (function (clustree clustree) clustree) clustree-make-joint))
(defun clustree-make-joint (ct1 ct2)
  "Make a joint clustering tree with two subtrees ct1 and ct2."
  (make-clustree :card (+ (clustree-card ct1)
                          (clustree-card ct2))
                 :adistr (adistr-add (clustree-adistr ct1)
                                     (clustree-adistr ct2))
                 :cdistr (cdistr-add (clustree-cdistr ct1)
                                     (clustree-cdistr ct2))
                 :subtrees (list ct1 ct2)))


(declaim (inline clustree-adjoin)
         (ftype (function (clustree clustree) clustree) clustree-adjoin))
(defun clustree-adjoin (ctree ct)
  "Join clustering tree ct to clustering tree ctree and return the result."
  (let ((ct1 (make-clustree :card (clustree-card ctree)
                            :adistr (clustree-adistr ctree)
                            :cdistr (clustree-cdistr ctree)
                            :subtrees (clustree-subtrees ctree))))
    (setf (clustree-card ctree)
          (+ (clustree-card ctree)
             (clustree-card ct)))
    (setf (clustree-adistr ctree)
          (adistr-add (clustree-adistr ctree)
                      (clustree-adistr ct)))
    (setf (clustree-cdistr ctree)
          (cdistr-add (clustree-cdistr ctree)
                      (clustree-cdistr ct)))
    (setf (clustree-subtrees ctree)
          (list ct1 ct))
    ctree))


(declaim (inline adistr-single)
         (ftype (function (cons) (vector (vector single))) adistr-single))
(defun adistr-single (i)
  "Return the attribute value distribution for single instance i."
  (map 'vector
       #'(lambda (a)
           (value-distribution (list i) a :type 'vector))
       (attributes-sequence)))


(declaim (inline cdistr-single)
         (ftype (function (cons) (vector fixnum)) cdistr-single))
(defun cdistr-single (i)
  "Return the category distribution for single instance i."
  (category-distribution (list i) :type 'vector))


(declaim (inline adistr-add)
         (ftype (function ((vector (vector fixnum))
                           (vector (vector fixnum)))
                          (vector (vector fixnum))) adistr-add))
(defun adistr-add (ad1 ad2)
  "Return the attribute value distribution obtained by adding distributions
ad1 and ad2."
  (map 'vector
       #'(lambda (vd1 vd2)
           (map 'vector
                #'(lambda (d1 d2)
                    (+ d1 d2))
                vd1 vd2))
       ad1 ad2))


(declaim (inline cdistr-add)
         (ftype (function ((vector fixnum)
                           (vector fixnum))
                          (vector fixnum)) cdistr-add))
(defun cdistr-add (cd1 cd2)
  "Return the category distribution obtained by adding distributions
cd1 and cd2."
  (map 'vector
       #'(lambda (d1 d2)
           (+ d1 d2))
       cd1 cd2))


(declaim (inline clustree-hosts)
         (ftype (function (clustree cons) list) clustree-hosts))
(defun clustree-hosts (ctree i)
  "Return conses consisting of subtrees of clustering tree ctree and
clustering utilities assuming instance i is added to them, sorted
best first."
  (sort (mapcar #'(lambda (st)
                    (cons st (clustree-try-add ctree st i)))
                (clustree-subtrees ctree))
        #'>
        :key #'(lambda (h) (rest h))))


;;; Clustering utility computation.


(declaim (ftype (function (clustree) real) clustree-utility))
(defun clustree-utility (ctree)
  "Compute the utility of clustering represented by the direct subcategories
of clustering tree ctree."
  (flet ((auxfn (tree)                  ; an auxiliary function
           (loop for a in (attributes-sequence)
                 as an from 0 below (attributes-length)
                 sum
                 (loop for vn from 0 below (attribute-values-length a)
                       sum (expt (/ (svref
                                     (svref (clustree-adistr tree) an)
                                     vn)
                                    (clustree-card tree))
                                 2)))))
    (let ((bias (auxfn ctree))
          (denom (length (clustree-subtrees ctree))))
      (/
       (loop for stree in (clustree-subtrees ctree)
             sum (* (/ (clustree-card stree) (clustree-card ctree))
                    (- (auxfn stree) bias)))
       denom))))


;;; Print the clustering.


(defun clustree-print (ctree stream depth &optional (indent 0))
  "Print the clustering tree ctree to stream."
  (format stream "~&~vT~A" indent (clustree-cdistr ctree))
  (dolist (st (clustree-subtrees ctree))
    (clustree-print st stream depth (+ indent 6))))
