;;;	$Id: cover.lsp,v 1.56 2000/02/10 11:59:02 pawel Exp $
;;; package COVER
;;; a simplified implementation of rule induction by
;;; sequential covering (roughly based on AQ and CN2)
;;; Copyleft 1998 Pawel Cichosz

(provide 'cover)
(require 'concept)
(require 'info)
(require 'util)

(defpackage "COVER"
    (:documentation "Package providing an implementation of rule induction
by sequential covering based on AQ and CN2.")
  (:use "COMMON-LISP" "CONCEPT" "INFO" "UTIL")
  (:export "*MAXSTAR*"
           "*CN2-DISJUNCTS*"
           "*CN2-THETA*"
           "LEARN"
           "CLASSIFY"
           "HYPOTHESIS-SIZE"))

(in-package "COVER")


(defconstant +cn2-maxval+ 0.0
  "The best possible CN2 complex evaluation.")

(defvar *atomic-star* nil
  "The set of all atomic complexes (with up to *cn2-disjuncts* values
per selector) used by CN2.")

(defvar *value-map* (make-hash-table :test #'eq)
  "The hash table mapping complexes to their evaluations, so as to
avoid recomputinmg them unnecesarily.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defparameter *maxstar* 1
  "The maximum star size for beem search.")

(defparameter *cn2-disjuncts* 1
  "The maximum number of number of attribute values in disjunctive selectors
for CN2.")

(defparameter *cn2-theta* 0.0
  "The statistical significance threshold for CN2.")


(declaim (inline learn)
         (ftype (function (list &key atom atom symbol) ruleset)
                learn))
(defun learn (instances &key
                        (target nil)
                        (default (majority-category instances))
                        (algorithm :aq))
  "Learn from instances assuming the domain described by
attributes and categories. Generate ruleset for category target
or for all categories if nil. Use the specified algorithm :aq/:cn2,
default :aq."
  (clrhash *value-map*)                 ; clear complex evaluations
  (when (eq algorithm :cn2)
    (setq *atomic-star* (atomic-star-generate)))
  (ruleset-generate instances
                    (instances-select-c instances target)
                    default
                    algorithm))

(declaim (ftype (function (ruleset cons &key) atom) classify))
(defun classify (rset i &key)
  "Return the category label of instance i predicted by ruleset rset."
  (dolist (r (ruleset-rules rset) (ruleset-default rset))
    (when (rule-covers r i)
      (return (rule-category r)))))


(defun hypothesis-size (rset &key)
  "Return the number of rules as the first value, the number of
non-universal selectors as the second value, and the number of
attribute values in selectors as the third value."
  (let ((nr 0)
        (ns 0)
        (nv 0))
    (dolist (r (ruleset-rules rset) (values nr ns nv))
      (incf nr)
      (multiple-value-bind (ns1 nv1)
          (rule-size r)
        (incf ns ns1)
        (incf nv nv1)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Hypotheses are represented by sets of rules.


(defstruct (ruleset
             (:print-function ruleset-print))
  "Rule set structure."
  (rules nil :type list)                ; list of rules
  (default nil :type atom))             ; default category


;;; A rule is represented by a complex and category label.


(defstruct (rule
             (:print-function rule-print))
  "Rule structure."
  (complex nil :type list)              ; cover
  (category nil :type atom))            ; category for covered instances


(declaim (inline rule-covers)
         (ftype (function (rule cons) t) rule-covers))
(defun rule-covers (r i)
  "Check whether rule r covers instance i."
  (complex-covers (rule-complex r) i))


(declaim (inline rule-size))
(defun rule-size (r)
  "Return the number of non-universal selectors in r as the first value
and the number of attribute values in selectors as the second value."
  (complex-size (rule-complex r)))


;;; A complex is a list of selectors or nil (if empty).


(declaim (inline complex-covers)
         (ftype (function (list cons) t) complex-covers))
(defun complex-covers (c i)
  "Check whether complex c covers instance i."
  (not (mismatch c (instance-values i)
		 :test #'selector-covers)))


(declaim (inline complex-more-specific)
         (ftype (function (list list) t) complex-more-specific))
(defun complex-more-specific (c1 c2)
  "Check whether complex c1 is more specific than complex c2."
  (or (null c1)                         ; empty complex is most specific
      (not (mismatch c1 c2
		     :test #'selector-more-specific))))


(declaim (inline complex-equiv)
         (ftype (function (list list) t) complex-equiv))
(defun complex-equiv (c1 c2)
  "Check whether complexes c1 and c2 are equivalent."
  (and (complex-more-specific c1 c2)
       (complex-more-specific c2 c1)))


(declaim (inline complex-intersect)
         (ftype (function (list list) list) complex-intersect))
(defun complex-intersect (c1 c2)
  "Return the intersection of comlexes c1 and c2."
  (let ((c (mapcar #'selector-intersect c1 c2))) ; intersect pairwise
    (unless (member nil c)              ; check if not empty
      c)))


(declaim (inline complex-covered)
         (ftype (function (list list) list) complex-covered))
(defun complex-covered (c instances)
  "Get the sublist of instances with only those covered by complex c."
  (select instances c
          :test #'(lambda (i c)
                    (complex-covers c i))))


(declaim (inline complex-uncovered)
         (ftype (function (list list) list) complex-uncovered))
(defun complex-uncovered (c instances)
  "Get the sublist of instances with only those not covered by complex c."
  (select instances c
          :test #'(lambda (i c)
                    (not (complex-covers c i)))))


(declaim (inline complex-count-covered)
         (ftype (function (list list) fixnum) complex-count-covered))
(defun complex-count-covered (c instances)
  "Count the number of instances covered by complex c."
  (count-if #'(lambda (i)
                (complex-covers c i))
            instances))


(declaim (inline complex-significant)
         (ftype (function (list list) t) complex-significant))
(defun complex-significant (c instances)
  "Check whenter complex c is statistically significant for instances."
  (let* ((covered (complex-covered c instances))
	 (n-cov (length covered))       ; number of instances covered by c
	 (n-all (length instances))     ; number of all instances
	 (c-cov (category-distribution covered)) ; distribution for covered
	 (c-all (category-distribution instances)) ; and for all instances
	 (d 0.0))                       ; distance metric between them
    (block significant
      (map nil
           #'(lambda (cc ca)            ; use counts for covered and all
               (unless (or (zerop cc) (zerop ca))
                 (let ((ec (* (/ n-cov n-all) ca))) ; expected cc
                   (incf d (* cc (log (/ cc ec))))))
               (when (> (* 2 d) *cn2-theta*)
                 (return-from significant t)))
           c-cov
           c-all))))


(declaim (inline complex-entropy)
         (ftype (function (list list) float) complex-entropy))
(defun complex-entropy (c instances)
  "Return the entropy of complex c for instances."
  (information (complex-covered c instances)))


(defun complex-size (c)
  "Return the number of non-universal selectors in c as the first value
and the number of attribute values in selectors as the second value."
  (let ((ns 0)
        (nv 0))
    (dolist (s c (values ns nv))
      (when (listp s)
        (incf ns)
        (incf nv (selector-size s))))))


;;; A selector is one of:
;;;   1. symbol :? -- covers anything,
;;;   2. list (v1 v2 v3...) -- covers values v1, v2, v3...,
;;;   3. nil -- covers nothing.


(declaim (inline selector-covers)
         (ftype (function ((or symbol list) atom) t) selector-covers))
(defun selector-covers (s v)
  "Check whether selector s covers attribute value v."
  (cond ((eq s :?) t)
	((listp s) (not (null (member v s))))
	(t nil)))


(declaim (inline selector-more-specific)
         (ftype (function ((or symbol list) (or symbol list)) t)
                selector-more-specific))
(defun selector-more-specific (s1 s2)
  "Check whether selector s1 is more specific that selector s2."
  (cond ((eq s2 :?) t)
	((eq s1 :?) nil)
	((listp s2) (subsetp s1 s2))
	(t nil)))


(declaim (inline selector-intersect)
         (ftype (function ((or symbol list) (or symbol list))
                          (or symbol list)) selector-intersect))
(defun selector-intersect (s1 s2)
  "Return the intersection of selectors s1 and s2."
  (cond ((eq s1 :?) s2)
	((eq s2 :?) s1)
	((and (listp s1) (listp s2))
	 (intersection s1 s2))
	(t nil)))


(declaim (inline selector-size))
(defun selector-size (s)
  "Return the number of of attribute values in s or 0 if s is empty or
universal."
  (cond ((listp s) (length s))
        (t 0)))


;;; The main thing begins here.


(declaim (ftype (function (list list atom symbol &optional list) ruleset)
                ruleset-generate))
(defun ruleset-generate (instances uncovered default algorithm &optional
                                   (rules nil))
  "Generate a list of rules from instances using algorithm (AQ/CN2)."
  (if (endp uncovered)                  ; if no uncovered instances
      (make-ruleset :rules (nreverse rules) ; return the reversed result
                    :default default)
      (let* ((c (complex-generate instances
                                  uncovered
                                  algorithm)) ; generate a new complex
             (covered (complex-covered c uncovered))) ; get its cover set
        (unless (null c)                ; if no inconsistency
          (ruleset-generate instances   ; recurse for more rules
                            (complex-uncovered c uncovered)
                            default
                            algorithm
                            (cons (make-rule
                                   :complex c
                                   :category (majority-category covered))
                                  rules))))))


(declaim (inline complex-generate)
         (ftype (function (list list symbol) list) complex-generate))
(defun complex-generate (instances uncovered algorithm)
  "Generate a complex for instances, given those yet uncovered, using
algorithm (AQ/CN2)."
  (case algorithm
    (:aq (complex-generate-aq instances uncovered))
    (:cn2 (complex-generate-cn2 uncovered))
    (otherwise (error "No such algorithm: ~A" algorithm))))


(declaim (ftype (function (list list) list) complex-generate-aq))
(defun complex-generate-aq (instances uncovered)
  "Generate a complex for instances, given those yet uncovered, using AQ."
  (let* ((seed (select-seed uncovered))
	 (cat (instance-category seed)))
    (complex-select-aq
     (star-generate-aq seed
                       (positive-instances uncovered cat)
                       (negative-instances instances cat)))))


(declaim (ftype (function (list list &optional list real list) list)
                complex-generate-cn2))
(defun complex-generate-cn2 (uncovered &optional
				       (c (atomic-complex)) ; best so far
				       (v (complex-evaluate-cn2 c uncovered))
				       (star (list c))) ; current star
  "Generate a complex for uncovered instances using CN2."
  (if (or (null star)                   ; if star is empty
	  (>= v +cn2-maxval+))          ; or we can't do any better
      c                                 ; return the best complex found
      (let* ((star1 (star-bound
                     (star-trim-cn2
                      (star-intersect star *atomic-star*) star)
                     uncovered
                     #'complex-evaluate-cn2))
             (c1 (complex-select-cn2 star1 uncovered))
             (v1 (heuristic-compute #'complex-evaluate-cn2 c1 uncovered)))
        (complex-generate-cn2
         uncovered
         (if (and c1 (> v1 v))          ; choose the better
             (progn (setq v v1) c1)
             c)
         v
         star1))))


(declaim (ftype (function (&optional fixnum (or atom list) atom) list)
                atomic-complex))
(defun atomic-complex (&optional (n -1) (vp nil) (vn nil))
  "Create an atomic complex with desired properties.
With no arguments, creates a universal complex.
If vp is a non-empty list, creates a complex with the nth selector vp.
Otherwise, creates a complex the nth selector of which covers
value vp but does not cover value vn."
  (let ((c nil))                        ; generated complex
    (do-attributes-n (a an (nreverse c))   ; for all attributes
      (if (= an n)
          (if (and vp (listp vp))     ; if vp is a list
              (push vp c)             ; use it as selector
              (let ((s                ; otherwise generate
                     (remove 
                      vn              ; selector not covering vn
                      (attribute-values a))))
                (if (and vp (not (member vp s))) ; check if
                    (return nil) ; it covers vp
                    (push s c))))     ; and use it, if so
          (push :? c)))))             ; this is not attribute number n


(declaim (inline atomic-star-generate)
         (ftype (function () list) atomic-star-generate))
(defun atomic-star-generate ()
  "Generate and return the list of all atomic complexes with up to
*cn2-disjuncts* values per selectors."
  (let ((astar nil))                    ; generated atomic star
    (do-attributes-n (a an astar)            ; complexes for all attributes
      (let* ((values (attribute-values a))
             (nv (length values)))      ; number of values of a
        (setq astar
              (nconc astar
                     (mapcar #'(lambda (s) ; for subsets of values
                                 (atomic-complex an s)) ; create complex
                             (nbutlast  ; skip empty subset
                              (subsets-n values
                                         (min (1- nv)
                                              *cn2-disjuncts*))))))))))


(declaim (ftype (function (cons list list &optional list) list)
                star-generate-aq))
(defun star-generate-aq (seed pos neg &optional
			      (star (list (atomic-complex))))
  "Generate the AQ star for seed, given pos (uncovered positive instances)
and neg (all negative instances) with respect to the category of seed."
  (let ((negseed (select-negseed star neg)))
    (if (null negseed)                  ; if no negative instances left
	star                            ; we've done it
        (let ((star1 (partial-star seed negseed)))
          (if (null star1)              ; if inconsistent data
              (star-generate-aq seed pos
                                (delete negseed neg) ; ignore inconsistency
                                star)   ; and try again
              (star-generate-aq seed pos neg
                                (star-bound
                                 (star-trim-aq
                                  (star-intersect star star1))
                                 pos
                                 #'complex-evaluate-aq)))))))


(declaim (ftype (function (cons cons) list) partial-star))
(defun partial-star (seed negseed)
  "Generate a list of all maximally general complexes covering seed
and not covering negseed."
  (let ((n 0)                           ; current attribute number
        (pstar nil))                    ; generated partial star
    (map nil
         #'(lambda (sv nv)              ; for a pair of seed/negseed values
             (let ((c (atomic-complex n sv nv)))
               (when c (push c pstar)))
             (incf n))
         (instance-values seed)
         (instance-values negseed))
    (nreverse pstar)))


(declaim (ftype (function (list list) list) star-intersect))
(defun star-intersect (star1 star2)
  "Return the intersection of two lists of complexes star1 and star2."
  (let ((star nil))
    (dolist (c1 star1 star)             ; for all complexes in star1
      (dolist (c2 star2)                ; and all complexes in star2
        (let ((c (complex-intersect c1 c2))) ; intersect them
          (unless (null c)              ; and if the result is not empty
            (push c star)))))))         ; add it to star


(declaim (inline positive-instances)
         (ftype (function (list atom) list) positive-instances))
(defun positive-instances (instances cat)
  "Get from instances those with category equal cat."
  (instances-select-c instances cat))


(declaim (inline negative-instances)
         (ftype (function (list atom) list) negative-instances))
(defun negative-instances (instances cat)
  "Get from instances those with category different from cat."
  (select instances cat
          :key #'instance-category
          :test (complement #'eql)))


(declaim (inline select-seed)
         (ftype (function (list) cons) select-seed))
(defun select-seed (instances)
  "Get the first of instances as seed."
  (first instances))


(declaim (inline select-negseed)
         (ftype (function (list list) cons) select-negseed))
(defun select-negseed (star instances)
  "Get the first of instances that is covered by any complex in star."
  (find-if #'(lambda (i)                ; check for instance i
               (some #'(lambda (c)      ; wether some complex c in star
                         (complex-covers c i)) ; covers it
                     star))
           instances))


(declaim (inline star-trim-aq)
         (ftype (function (list) list) star-trim-aq))
(defun star-trim-aq (star)
  "Trim star by removing any complexes more specific than some other
complexes."
  (delete-if #'(lambda (c1)             ; check for each complex c1
                 (some #'(lambda (c2)   ; whether some
                           (and (not (eq c1 c2)) ; other complex c2 is more
                                (complex-more-specific c1 c2))) ; specific
                       star))
             star))


(declaim (inline star-trim-cn2)
         (ftype (function (list list) list) star-trim-cn2))
(defun star-trim-cn2 (star star1)
  "Trim star by removing any duplicate complexes or complexes that are
in star1."
  (nset-difference
   (delete-duplicates star :test #'complex-equiv)
   star1                                ; exclude complexes present in star1
   :test #'complex-equiv))


(declaim (inline star-bound)
         (ftype (function (list list (function (list list) real)) list)
                star-trim-aq))         
(defun star-bound (star instances heuristic)
  "Return at most *maxstar* best complexes from star. Evaluates complexes
by calling heuristic for instances. The returned bounded star is sorted
\(best first)."
  (setq star
	(stable-sort star
		     #'>
		     :key #'(lambda (c)
			      (heuristic-compute heuristic c instances))))
  (nbutlast star (max 0 (- (length star) *maxstar*)))) ; take first *maxstar*


(declaim (inline heuristic-compute)
         (ftype (function ((function (list list) real) list instances) real)
                heuristic-compute))
(defun heuristic-compute (heuristic c instances)
  "Get the value of heuristic for complex c with respect to instances.
The value is stored as a complex-indexed value in hash table *value-map*
and computed only when necessary."
  (or (gethash c *value-map*)           ; eq complexes are always
      (setf (gethash c *value-map*)     ; evaluated w.r.t. same
	    (funcall heuristic c instances)))) ; instance sets


(declaim (inline complex-evaluate-aq)
         (ftype (function (list list) real) complex-evaluate-aq))
(defun complex-evaluate-aq (c instances)
  "Evaluate complex c with respect to instances. Returns the number of
instances covered by c."
  (complex-count-covered c instances))


(declaim (inline complex-evaluate-cn2)
         (ftype (function (list list) real) complex-evaluate-cn2))
(defun complex-evaluate-cn2 (c instances)
  "Evaluate complex c with respect to instances. Returns the negated
\(to minimize) entropy of complex c."
  (- (complex-entropy c instances)))


(declaim (inline complex-select-aq)
         (ftype (function (list) list) complex-select-aq))
(defun complex-select-aq (star)
  "Select the first complex from star."
  (first star))


(declaim (inline complex-select-cn2)
         (ftype (function (list list) list) complex-select-cn2))
(defun complex-select-cn2 (star instances)
  "Select the first complex from star statistically significant on instances."
  (find-if #'(lambda (c)
               (complex-significant c instances))
           star))


;;; Printing.


(defun ruleset-print (rset stream depth)
  "Print set of rules rset."
  (declare (ignore depth))              ; not using this argument
  (format stream "~&Rules:~%")
  (let ((i 1))                          ; rule number
    (dolist (r (ruleset-rules rset))
      (format stream "~&~A. ~A~%" i r)
      (incf i)))
  (format stream "~&Default category: ~A~%" (ruleset-default rset)))
  
(defun rule-print (r stream depth)
  "Print rule r."
  (declare (ignore depth))              ; not using this argument
  (format stream "IF < ")
  (map nil
       #'(lambda (a s)
	   (unless (eq s :?)
             (format stream "~A=( " (attribute-name a))
             (dolist (v s)
               (format stream "~A " v))
             (format stream ") ")))
       (attributes-sequence)            ; for all attributes
       (rule-complex r))                ; and corresponding selectors
  (format stream ">~%~3TTHEN ~A" (rule-category r)))
