;;;	$Id: discretize.lsp,v 1.21 2000/02/10 11:19:43 pawel Exp $	
;;; package DISCRETIZE
;;; utilities for the discretization of continuous attributes
;;; Copyleft 1998 Pawel Cichosz

(provide 'discretize)
(require 'concept)
(require 'util)

(defpackage "DISCRETIZE"
    (:documentation "Package providing functions for the discrtetization of
continuous attributes in concept learning data sets.")
  (:use "COMMON-LISP" "CONCEPT" "UTIL")
  (:export "*CHI-THETA*"
           "EQUAL-WIDTH"
           "EQUAL-FREQUENCY"
           "CHI-MERGE"
           "ATTRIBUTES-DISCRETIZE"
           "INSTANCES-DISCRETIZE"))

(in-package "DISCRETIZE")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defparameter *chi-theta* 0.9
  "The significance threshold for ChiMerge discretization.")


(declaim (ftype (function (list &key (or fixnum list) t symbol) sequence)
                attributes-discretize))
(defun attributes-discretize (instances &key
                                        (intervals 5)
                                        (extend nil)
                                        (algorithm :equal-width))
  "Discretize continuous attributes for the given set of instances, using
the specified algorithm (:equal-width, :equal-frequency, :chi-merge).
The interpretation of number or sequence of numbers intervals (must
be the same length as attributes) is algorithm-dependent: it specifies
either the exact number of discretization intervals or an upper/lower
bound thereof. If extend is true, two special intervals are added
to the created interval lists (if necessary) for all values less/greater
than the lowest/highest value observed in instances. Modifies attributes
maintained by CONCEPT appropriately and returns them. Instances are
not modified in any way. The discrete values correspond to subranges of
original continuous values and are represented by symbols created from
strings \"v1:v2\", where v1 and v2 are subrange boundaries."
  (map-into (attributes-sequence)
            #'(lambda (a n1)
                (if (attribute-continuous-p a)
                    (attribute-discretize a instances n1 extend algorithm)
                    a))
            (attributes-sequence)
            (if (listp intervals)
                intervals
                (fill (make-sequence 'list (attributes-length))
                      intervals))))


(declaim (ftype (function (list &key) list) instances-discretize))
(defun instances-discretize (instances &key)
  "Discretize attribute values in the given set of instances,
assuming attributes have been discretized already. The list of instances
is modified and returned."
  (map-into instances
            #'(lambda (i)
                (do-attributes (a i)
                  (instance-discretize i a)))
            instances))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defconstant *interval-separator* #\:
  "Character used to separate interval boundaries in discretized values.")


;;; Discretization algorithms use (internally) intervals.
;;; Interval boundary is either a number or nil (denoting +/- infinity).


(defstruct (interval
             (:print-function interval-print))
  (low 0.0 :type (or real null))        ; lower boundary
  (high 0.0 :type (or real null))       ; higher boundary
  (distrib nil :type list)              ; category distribution (optional)
  (card 0 :type integer))               ; number of instances (optional)


;;; Still some interfacing functions.


(declaim (ftype (function (cons list fixnum t symbol) cons)
                attribute-discretize))
(defun attribute-discretize (a instances n extend algorithm)
  "Discretize continuos attribute a in the given instances set
into (exactly/at most/at least) intervals using the specified algorithm.
Returns the discretized attribute."
  (create-attribute (attribute-name a)
                    (values-discretize
                     (mapcar #'(lambda (i)
                                 (cons (instance-value-of i (attribute-name a))
                                       (instance-category i)))
                             instances)
                     n extend algorithm)))


(declaim (ftype (function (list fixnum t symbol) list) values-discretize))
(defun values-discretize (vcs n extend algorithm)
  "Discretize continuous values in the list vcs of (value . category)
conses into (exactly/at most/at least) intervals using
algorithm. Return the list of created discrete values, corresponding to
subranges of original continuous values, represented by symbols created
from strings \"v1:v2\", where v1 and v2 are subrange boundaries. If extend
is true, the first and last intervals always range from/to +/- infinity
represented by nil."
  (if (null vcs)                        ; if no values to discretize
      (make-discrete nil nil)           ; create one interval
      (let* ((ovcs (sort vcs #'<        ; ordered value-category list
                         :key #'(lambda (vc) (first vc))))
             (intvs (case algorithm     ; discretized values
                      (:equal-width (equal-width-discretize ovcs n))
                      (:equal-frequency (equal-frequency-discretize ovcs n))
                      (:chi-merge (chi-merge-discretize ovcs n))
                      (otherwise (error "No such algorithm: ~A" algorithm))))
             (l (interval-low (first intvs))) ; the lowest boundary
             (h (interval-high (first (last intvs))))) ; the highest boundary
        (if (and (null l) (null h))
            (make-discrete nil nil)
            (nconc (unless (or (not extend) (null l))
                     (list (make-discrete nil l))) ; extra left interval
                   (mapcar #'(lambda (i)
                               (make-discrete (interval-low i)
                                              (interval-high i)))
                           intvs)
                   (unless (or (not extend) (null h))
                     (list (make-discrete h nil)))))))) ; extra right interval


(declaim (ftype (function (cons cons) cons) instance-discretize))
(defun instance-discretize (i a)
  "Replace the continuous value of attribute a in instance i by
the corresponding discrete value and return the modified instance."
  (let ((vd (discrete-value (instance-value-of i (attribute-name a))
			    (attribute-values a))))
    (when vd
      (setf (instance-value-of i (attribute-name a)) vd))
    i))


;;; The actual discretization algorithms.


(declaim (ftype (function (list fixnum) list) equal-width-discretize))
(defun equal-width-discretize (vcs n)
  "Discretize continuous values in the ordered list vcs of
\(value . category) conses into n intervals of equal width. Two extra
intervals are added, for values less/greater than the lowest/highest
value in vcs. Return the list of created intervals."
  (loop with d = (/ (- (first (first (last vcs))) ; margin
                       (first (first vcs))) (* 2 n))
        with low = (- (first (first vcs)) d)
        with high = (+ (first (first (last vcs))) d)
        with delta = (/ (- high low) n)
        for l from low below high by delta
        as h = (+ l delta)
        collect (make-interval :low l :high h))) ; collect this interval


(declaim (ftype (function (list fixnum) list) equal-frequency-discretize))
(defun equal-frequency-discretize (vcs n)
  "Discretize continuous values in the ordered list vcs of
\(value . category) conses into n intervals with equal number of
vcs elements in each. Return the list of created intervals."
  (loop with d = (/ (- (first (first (last vcs))) ; margin
                       (first (first vcs))) (* 2 n))
        with vcp = (partition vcs n)    ; partition values
        for ps on vcp
        as p = (first ps)               ; subset for current interval
        as p1 = (second ps)             ; subset for next interval
        as h = (if p1                   ; higher boundary
                   (/ (+ (first (first p1)) (first (first (last p)))) 2)
                   (+ (first (first (last p))) d))
        and as l = (- (first (first p)) d) then h ; lower boundary
        when (not (= l h))              ; if there is a real interval
        collect (make-interval :low l :high h ; collect it
                               :card (length p)
                               :distrib
                               (frequency-distribution
                                vcs (categories-sequence)
                                :key #'(lambda (vc)
                                         (let ((v (first vc))
                                               (c (rest vc)))
                                           (when (and (>= v l)
                                                      (< v h))
                                             c)))))))


(declaim (ftype (function (list fixnum &optional list float) list)
                chi-merge-discretize))
(defun chi-merge-discretize (vcs n &optional
                                 (intvs (equal-frequency-discretize
                                         vcs
                                         (length vcs)))
                                 (chimin 0.0))
  "Discretize continuous values in the ordered list vcs of
\(value . category) conses into at least n intervals using the ChiMerge
algorithm. Return the list of  created intervals."
  (if (or (<= (length intvs) n)         ; if no more than n intervals
          (>= chimin *chi-theta*))      ; or too significant differences
      intvs                             ; return them
      (let ((in 0)                      ; current interval number
            (imin 0)                    ; minimum chi-square interval number
            (chimin most-positive-single-float)) ; minimum chi-square value
        (map nil                        ; find intervals to join
             #'(lambda (i1 i2)
                 (let ((chi (chi-square i1 i2)))
                   (when (< chi chimin) ; new minimum
                     (setq chimin chi imin in))) ; store it
                 (incf in))             ; next interval number
             intvs
             (rest intvs))
        (chi-merge-discretize vcs n (intervals-join intvs imin) chimin))))


(declaim (ftype (function (interval interval) float) chi-square))
(defun chi-square (i1 i2)
  "Compute the chi-square statistic for intervals i1, i2."
  (let* ((c-i1 (interval-distrib i1))   ; category distribution for i2
         (c-i2 (interval-distrib i2))   ; category distribution for i1
         (c-i1i2 (mapcar #'+ c-i1 c-i2)) ; category distribution for i1+i2
         (n-i1 (interval-card i1))      ; number of values in i1
         (n-i2 (interval-card i2))      ; number of values in i2
         (n-i1i2 (+ n-i1 n-i2))         ; number of values in i1+i2
         (chi 0.0))                     ; chi-square value
    (map nil
         #'(lambda (c1 c2 c12)          ; use counts for i1, i2, i1+i2
             (unless (zerop c12)
               (let ((e1 (* (/ n-i1 n-i1i2) c12)) ; expected count for i1
                     (e2 (* (/ n-i2 n-i1i2) c12))) ; expected count for i2
                 (unless (zerop e1)
                   (incf chi (/ (expt (- c1 e1) 2) e1)))
                 (unless (zerop e2)
                   (incf chi (/ (expt (- c2 e2) 2) e2))))))
         c-i1 c-i2 c-i1i2)
    chi))


;;; Some utility functions.


(declaim (inline make-discrete)
         (ftype (function ((or real null) (or real null)) symbol)
                make-discrete))
(defun make-discrete (l h)
  "Make a discrete value form interval [l, h)."
  (make-symbol (format nil "~A~A~A" l *interval-separator* h)))


(declaim (inline discrete-low)
         (ftype (function (symbol) (or real null)) discrete-low))
(defun discrete-low (v)
  "Get the lower boundary corresponding to discretized value v."
  (let ((s (symbol-name v)))
    (nth-value 0 (read-from-string
                  (subseq s 0 (position *interval-separator* s))))))


(declaim (inline discrete-high)
         (ftype (function (symbol) (or real null)) discrete-high))
(defun discrete-high (v)
  "Get the higher boundary corresponding to discretized value v."
  (let ((s (symbol-name v)))
    (nth-value 0 (read-from-string
                  (subseq s (1+ (position *interval-separator* s)))))))


(declaim (inline discrete-value)
         (ftype (function (real list) (or symbol null)) discrete-low))
(defun discrete-value (v vds)
  "Get the element of discretized values vds for continuous value v."
  (when (numberp v)
    (find-if #'(lambda (vd)
                 (discrete-inside vd v))
             vds)))


(declaim (inline discrete-inside)
         (ftype (function (symbol real) t) discrete-inside))
(defun discrete-inside (vd v)
  "Check whether value v is inside the interval corresponding to
discretized value vd."
  (let ((l (discrete-low vd))
        (h (discrete-high vd)))
    (and (or (null l) (>= v l))
         (or (null h) (< v h)))))


(declaim (inline interval-inside)
         (ftype (function (interval real) t) interval-inside))
(defun interval-inside (i v)
  "Check whether value v is inside interval i."
  (let ((l (interval-low i))
        (h (interval-high i)))
    (and (or (null l) (>= v l))
         (or (null h) (< v h)))))


(declaim (ftype (function (list fixnum) list) intervals-join))
(defun intervals-join (intvs n)
  "Join intervals number n and n+1 from intvs and return the modified
list of intervals."
  (let* ((ni (length intvs))
         (tail (nthcdr n intvs))
         (head (nbutlast intvs (- ni n)))
         (i1 (first tail))              ; the two intervals to join
         (i2 (second tail)))
    (nconc head
           (list (make-interval :low (interval-low i1) ; join them
                                :high (interval-high i2)
                                :card (+ (interval-card i1)
                                         (interval-card i2))
                                :distrib (mapcar #'+
                                                 (interval-distrib i1)
                                                 (interval-distrib i2))))
           (nthcdr 2 tail))))


;;; Printing.


(declaim (inline interval-print))
(defun interval-print (intv stream depth)
  (declare (ignore depth))
  (format stream "[~A,~A)" (interval-low intv) (interval-high intv)))
