;;;	$Id: missing.lsp,v 1.4 2000/02/10 11:20:39 pawel Exp $	
;;; package MISSING
;;; most simple utilities for handling missing attribute values
;;; Copyleft 1998 Pawel Cichosz

(provide 'missing)
(require 'concept)
(require 'util)

(defpackage "MISSING"
    (:documentation "Package providing functions for handling missing
attribute values.")
  (:use "COMMON-LISP" "CONCEPT" "UTIL")
  (:export "MOST-COMMON"
	   "CATEGORY-MOST-COMMON"
           "MISSING-ELIMINATE"
           "MISSING-FILL"))

(in-package "MISSING")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (inline missing-eliminate)
	 (ftype (function (list &key) list) missing-eliminate))
(defun missing-eliminate (instances &key)
  "Eliminate non-destructively instances with missing attribute values."
  (select instances nil
          :key #'(lambda (i)
		   (some #'value-unknown-p
			 (instance-values i)))))


(declaim (inline missing-fill)
	 (ftype (function (list &key symbol) list) missing-fill))
(defun missing-fill (instances &key algorithm)
  "Fill missing attribute values for instances using the specified algorithm
\(:most-common, :category-most-common). The instances list is modified and
returned."
  (case algorithm
    (:most-common (most-common-fill instances))
    (:category-most-common (category-most-common-fill instances))
    (otherwise (error "No such algorithm: ~A" algorithm))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(declaim (ftype (function (list) list) most-common-fill))
(defun most-common-fill (instances)
  "Fill missing attribute values for instances with most common values.
The instances list is modified and returned."
  (let* ((values (mapcar #'(lambda (a)	; values for filling
			     (majority-value instances a))
			 (attributes-sequence))))
    (map-into instances
	      #'(lambda (i)
		  (instance-fill i values))
	      instances)))


(declaim (ftype (function (list) list) most-common-fill))
(defun category-most-common-fill (instances)
  "Fill missing attribute values for instances with most common values
in their categories. The instances list is modified and returned."
  (let* ((values (mapcar #'(lambda (c)	; values for filling
			     (mapcar #'(lambda (a)
					 (majority-value-c instances a c))
				     (attributes-sequence)))
			 (categories-sequence))))
    (map-into instances
	      #'(lambda (i)
		  (instance-fill i (elt values
					(category-position
					 (instance-category i)))))
	      instances)))


(declaim (ftype (function (cons sequence) cons) instance-most-common-fill))
(defun instance-fill (i values)
  "Fill missing attribute values in instance i with values and return
the modified instance."
  (let ((an 0))				; current attribute number
    (map nil
	 #'(lambda (v v1)
	     (when (value-unknown-p v)
	       (setf (instance-value-at i an) v1))
	     (incf an))
	 (instance-values i)
	 values)
    i))
