;;;	$Id: tdidt.lsp,v 1.50 1998/09/19 09:11:07 pawel Exp $
;;; package TDIDT
;;; a simplified implementation of ID3-like top-down induction of
;;; decision trees
;;; Copyleft 1998 Pawel Cichosz

(provide 'tdidt)
(require 'concept)
(require 'info)
(require 'util)

(defpackage "TDIDT"
    (:documentation "Package providing an implementation of ID3-like
top-down induction of decision trees.")
  (:use "COMMON-LISP" "CONCEPT" "INFO" "UTIL")
  (:export "LEARN"
           "CLASSIFY"
           "HYPOTHESIS-SIZE"))

(in-package "TDIDT")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The public interface.                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(declaim (inline learn)
         (ftype (function (list &key) dectree) learn))
(defun learn (instances &key)
  "Learn from instances assuming the domain described by
attributes and categories."
  (dectree-build instances
                 (let ((tests nil))
                   (dotimes (n (attributes-length)
                             tests)     ; take all attribute numbers
                     (setq tests (nconc tests (list n))))))) ; as tests


(declaim (ftype (function (dectree cons &key) atom) classify))
(defun classify (tree i &key)
  "Return the category label of instance i predicted by tree."
  (if (dectree-leaf-p tree)             ; if just a leaf
      (dectree-category tree)           ; return its category label
      (classify (decbranch-subtree      ; otherwise
                 (dectree-branch        ; use the proper subtree
                  tree
                  (instance-value-of i (dectree-test tree))))
                i)))


(defun hypothesis-size (tree &key)
  "Return the number of nodes of tree as the first value and the number
of leaves as the second value."
  (cond ((dectree-leaf-p tree)
         (values 0 1))
        (t (let ((nn 1)      ; number of nodes, including this one
                 (nl 0))     ; number of leaves
             (dolist (b (dectree-branches tree) (values nn nl))
               (multiple-value-bind (nn1 nl1)
                   (hypothesis-size (decbranch-subtree b))
                 (incf nn nn1)
                 (incf nl nl1)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The implementation.                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; A tree is represented by structure consisting of either a test and
;;; branches corresponding to its results or a category label.


(defstruct (dectree
             (:print-function dectree-print))
  "Decision tree structure."
  (category nil :type atom)             ; either a category label
  (test nil :type symbol)               ; or a test attribute name
  (branches nil :type list))            ; and a list of branches


;;; A decision tree branch consists of a test value and the corresponding
;;; subtree.


(defstruct (decbranch
             (:print-function decbranch-print))
  "Decision tree branch structure."
  (value nil :type atom)                ; the test value for this branch
  (subtree nil :type dectree))          ; the corresponding subtree


;;; A tree leaf is represented by a category label (symbol).


(declaim (inline dectree-leaf-p)
         (ftype (function (dectree) t) dectree-leaf-p))
(defun dectree-leaf-p (tree)
  "Check whether tree is actually a leaf."
  (null (dectree-branches tree)))


;;; A tree node is represented by a test attribute name and a list of
;;; branches.


(declaim (inline dectree-branch)
         (ftype (function (dectree attribute-value) decbranch)
                dectree-branch))
(defun dectree-branch (tree v)
  "Get the branch going out of the root of tree corresponding to
test result v."
  (find-if #'(lambda (b)
               (eql (decbranch-value b) v))
           (dectree-branches tree)))


;;; Here comes the main thing -- the tree construction algorithm.


(declaim (ftype (function (list list &optional atom) dectree)
                dectree-build))
(defun dectree-build (instances tests &optional
				(default (category-at 0)))
  "Build a decision tree for instances.
Consider attributes whose names are listed in tests for splitting.
Use default as default category label."
  (let* ((distrib (category-distribution instances)) ; category distribution
	 (major (category-at (argmax distrib)))) ; majority category
    (cond ((endp instances)             ; if no instances
	   (make-dectree :category default)) ; -- use default label
	  ((or (= (count 0 distrib :test #'<) 1) ; if single category
               (endp tests))            ; or no tests left
	   (make-dectree :category major)) ; -- use majority category
	  (t                            ; otherwise do the hard job
	   (let* ((tn                   ; select split attribute number
                   (select-test tests instances))
		  (ta (attribute-at tn))) ; test attribute
	     (make-dectree
              :test (attribute-name ta) ; test selected
              :branches                 ; and branch to subtrees
              (map 'list
                   #'(lambda (v)        ; for each value
                       (make-decbranch  ; create a branch
                        :value v        ; that for value v leads to
                        :subtree
                        (dectree-build  ; recursively built subtree
                         (instances-select-v instances tn v)
                         (remove tn tests) ; remove used test
                         major)))       ; default to majority category
                   (attribute-values ta))))))))


;;; The most tedious part: split attribute selection.


(declaim (inline select-test)
         (ftype (function (list list) fixnum) select-test))
(defun select-test (tests instances)
  "Select an attribute number from tests maximizing the information gain
for instances (standard ID3 heuristic)."
  (let ((gains (map 'vector             ; information gains for each test
                    #'(lambda (n)
                        (info-gain n instances)) ; we could only use entropy
                    tests)))
    (elt tests (argmax gains))))        ; attribute with maximum gain


(declaim (inline info-gain)
         (ftype (function (fixnum list) float) info-gain))
(defun info-gain (n instances)
  "Compute the information gain of attribute number n for instances."
  (- (information instances)
     (entropy n instances)))


;;; Printing.


(defun dectree-print (tree stream depth &optional (indent 0))
  "Print tree in a readable form to stream."
  (cond ((dectree-leaf-p tree)          ; if just a leaf
	 (format stream "~&~vT[~A]" indent
		 (dectree-category tree))) ; print its category label
	(t (format stream "~&~vT<~A>" indent ; otherwise
                   (dectree-test tree)) ; print test name
           (dolist (b (dectree-branches tree)) ; and all the branches
             (decbranch-print b stream depth indent)))))

(defun decbranch-print (branch stream depth &optional (indent 0))
  (format stream "~&~vT =~A" indent (decbranch-value branch))
  (dectree-print (decbranch-subtree branch) ; and recurse to subtrees
                 stream
                 depth
                 (+ indent 6)))
