;;;	$Id: tester.lsp,v 1.18 1998/09/19 11:59:14 pawel Exp $	
;;; package TESTER
;;; simple functionality for testing learning algorithms
;;; Copyleft 1998 Pawel Cichosz

(provide 'tester)
(require 'concept)
(require 'util)
(require 'discretize)
(require 'missing)


(defpackage "TESTER"
    (:documentation "Package providing basic functionality for testing
learning algorithms.")
  (:use "COMMON-LISP" "CONCEPT" "UTIL")
  (:export "TRAIN"
           "CLASSIFY"
           "HYPOTHESIS-SIZE"
           "TEST"
           "CROSS-VALIDATE"
           "K-FOLD-CROSS-VALIDATE"
           "DISCRETIZE-ATTRIBUTES"
           "DISCRETIZE-INSTANCES"
           "DISCRETIZE-ALL"
           "MISSING-ELIMINATE"
           "MISSING-FILL"
           "EXPERIMENT"))

(in-package "TESTER")


(declaim (inline train))
(defun train (instances &rest options &key
                        (attributes (attributes-sequence))
                        (categories (categories-sequence))
                        (package "TDIDT")
                        &allow-other-keys)
  "Train the algorithm provided by package from instances assuming
  the domain described by attributes and categories."
  (concept-setup attributes categories)
  (apply (find-symbol "LEARN" package)
         `(,instances :allow-other-keys t ,@options)))


(declaim (inline classify))
(defun classify (hypothesis i &rest options &key
                            (attributes (attributes-sequence))
                            (categories (categories-sequence))
                            (package "TDIDT")
                            &allow-other-keys)
  "Return the category label of instance i predicted by hypothesis
using the algorithm provided by package."
  (concept-setup attributes categories)
  (apply (find-symbol "CLASSIFY" package)
         `(,hypothesis ,i :allow-other-keys t ,@options)))


(declaim (inline hypothesis-size))
(defun hypothesis-size (hypothesis &rest options &key
                                   (attributes (attributes-sequence))
                                   (categories (categories-sequence))
                                   (package "TDIDT")
                                   &allow-other-keys)
  "Return the hypothesis size for package."
  (concept-setup attributes categories)
  (apply (find-symbol "HYPOTHESIS-SIZE" package)
         `(,hypothesis :allow-other-keys t ,@options)))


(defun test (hypothesis instances &rest options &key &allow-other-keys)
  "Return the sample error of hypothesis for instances.
Assume attributes and categories describe the domain."
  (if (endp instances)
      0
      (float (/ (count-if
                 #'(lambda (i)          ; divide the number of mistakes
                     (not (eq (apply #'classify
                                     `(,hypothesis ,i ,@options))
                              (instance-category i))))
                 instances)
                (length instances)))))  ; by the number of all instances


(defun cross-validate (instances r &rest options &key &allow-other-keys)
  "Perform cross-validation by training on a random subset of instances
of size equal r times the whole set size and then test on the remaining
instances, assuming the domain described by attributes and categories.
Returns the testing error as the first value, the training error
as the second value, and the hypothesis size as the third value."
  (let* ((isets (random-bipartition instances r)) ; make two subsets
         (train-set (first isets))
         (test-set (second isets))
         (h (apply #'train              ; generate hypothesis
                   `(,train-set ,@options)))
         (s (multiple-value-list (apply #'hypothesis-size `(,h ,@options))))
         (train-error (apply #'test     ; error on training set
                             `(,h ,train-set ,@options)))
         (test-error (apply #'test      ; error on testing set
                            `(,h ,test-set ,@options))))
    (values test-error train-error s)))


(defun k-fold-cross-validate (instances k &rest options &key
                                        &allow-other-keys)
  "Perform k-fold cross-validation, assuming the domain described by
attributes and categories. Returns the average testing error as
the first value and a list of lists containing all k testing errors
and corresponding hypothesis sizes as the second value."
  (let* ((test-sets (random-partition instances k))
         (results (mapcar
                   #'(lambda (ts)       ; for each testing set
                       (let* ((tr (reduce ; generate training set
                                   #'(lambda (s1 s2)
                                       (cond ((eq s1 ts) s2)
                                             ((eq s2 ts) s1)
                                             (t (append s1 s2))))
                                   test-sets
                                   :initial-value nil))
                              (h (apply #'train ; learn a hypothesis
                                        `(,tr ,@options)))
                              (s  (multiple-value-list ; hypothesis size
                                      (apply #'hypothesis-size
                                             `(,h ,@options))))
                              (tse (apply #'test ; testing error
                                          `(,h ,ts ,@options))))
                         (cons tse (list s))))
                   test-sets)))
    (values (avg results :key #'car) results)))


(declaim (inline discretize-attributes))
(defun discretize-attributes (instances &rest options &key
                                        (attributes (attributes-sequence))
                                        (categories (categories-sequence))
                                        &allow-other-keys)
  "Discretize continuous attributes for given instances and concept domain.
Modifies and returns attributes."
  (concept-setup attributes categories)
  (apply #'discretize:attributes-discretize
         `(,instances :allow-other-keys t ,@options)))


(declaim (inline discretize-instances))
(defun discretize-instances (instances &rest options &key
                                       (attributes (attributes-sequence))
                                       (categories (categories-sequence))
                                       &allow-other-keys)
  "Discretize continuous attribute values for given instances.
Modifies and returns instances."
  (concept-setup attributes categories)
  (apply #'discretize:instances-discretize
         `(,instances :allow-other-keys t ,@options)))


(declaim (inline discretize-all))
(defun discretize-all (instances &rest options &key &allow-other-keys)
  "Discretize continuous attributes for given instances and concept domain.
Modifies both instances and attributes, returnes modified instances."
  (apply #'discretize-attributes
         `(,instances :allow-other-keys t ,@options))
  (apply #'discretize-instances
         `(,instances :allow-other-keys t ,@options)))


(declaim (inline missing-eliminate))
(defun missing-eliminate (instances &rest options &key
                                    (attributes (attributes-sequence))
                                    (categories (categories-sequence))
                                    &allow-other-keys)
  "Eliminate instances with missing attribute values. Modifies and returns
instances."
  (concept-setup attributes categories)
  (apply #'missing:missing-eliminate
         `(,instances :allow-other-keys t ,@options)))


(declaim (inline missing-fill))
(defun missing-fill (instances &rest options &key
                               (attributes (attributes-sequence))
                               (categories (categories-sequence))
                               &allow-other-keys)
  "Fill missing attribute values in instances. Modifies and returns
instances."
  (concept-setup attributes categories)
  (apply #'missing:missing-fill
         `(,instances :allow-other-keys t ,@options)))


(defun experiment (testfn testargs algorithms &rest options
                          &key &allow-other-keys)
  "Run testfn for algorithms and return results. Each of algorithms may
be a symbol -- passed to testfn as :package -- or list
-- the head of which is passed to testfn as :package and the rest as
options. Returns assoc-list in which each of algorithms is associated with
the corresponding result of testfn."
  (mapcar #'(lambda (alg)
              (let ((pkg (if (atom alg) alg (first alg))) ; package
                    (opt (when (listp alg) (rest alg)))) ; specific options
                (cons alg (apply testfn
                                 `(,@testargs :package ,pkg
                                   ,@options ,@opt)))))
          algorithms))
